//@author: tonfilm
//@help: recursively filles closed contours
//@tags:
//@credits:

// --------------------------------------------------------------------------------------------------
// PARAMETERS:
// --------------------------------------------------------------------------------------------------

//transforms
float4x4 tW: WORLD;        //the models world matrix
float4x4 tV: VIEW;         //view matrix as set via Renderer (EX9)
float4x4 tP: PROJECTION;
float4x4 tWVP: WORLDVIEWPROJECTION;

//texture
texture Tex <string uiname="Texture";>;
sampler Samp = sampler_state    //sampler for doing the texture-lookup
{
    Texture   = (Tex);          //apply a texture to the sampler
    MipFilter = point;         //sampler states
    MinFilter = point;
    MagFilter = point;
};

//texture transformation marked with semantic TEXTUREMATRIX to achieve symmetric transformations
float4x4 tTex: TEXTUREMATRIX <string uiname="Texture Transform";>;

//the data structure: "vertexshader to pixelshader"
//used as output data with the VS function
//and as input data with the PS function
struct vs2ps
{
    float4 Pos  : POSITION;
    float2 TexCd : TEXCOORD0;
};

// --------------------------------------------------------------------------------------------------
// VERTEXSHADERS
// --------------------------------------------------------------------------------------------------
vs2ps VS(
    float4 PosO  : POSITION,
    float4 TexCd : TEXCOORD0)
{
    //declare output struct
    vs2ps Out;

    //transform position
    Out.Pos = mul(PosO, tWVP);
    
    //transform texturecoordinates
    Out.TexCd = mul(TexCd, tTex);

    return Out;
}

// --------------------------------------------------------------------------------------------------
// PIXELSHADERS:
// --------------------------------------------------------------------------------------------------

float2 PixSize;
float2 StartPos;

float4 FillCol : COLOR;

struct Neighbours4
{
	float4 n[4];
};

Neighbours4 GetNeighbours(sampler tex, float2 pos, float2 pix)
{
	Neighbours4 n;
	
	n.n[0] = tex2D(tex, pos - float2(0, pix.y));
	n.n[1] = tex2D(tex, pos - float2(pix.x, 0));
	n.n[2] = tex2D(tex, pos + float2(0, pix.y));
	n.n[3] = tex2D(tex, pos + float2(pix.x, 0));
	
	return n;
}

float Epsilon;

bool Equals(float4 a, float4 b)
{
	return length(a.rgb-b.rgb) < Epsilon;
}

float4 PS(vs2ps In): COLOR
{
    float4 col = tex2D(Samp, In.TexCd);
	
	//paint start pixel if fill color active
	if(FillCol.a > 0.5)
	{
		if(all(abs(In.TexCd - StartPos) < PixSize*0.5))
		{
			col = FillCol;
		}
	}
	
	//check if pixel is active
	if(col.a > 0.5)
	{
		//set fill col
		col = FillCol;
	}
	else //check if it should get the fill color
	{
		//check if any neighbour is active
		Neighbours4 n = GetNeighbours(Samp, In.TexCd, PixSize);
		for(int i=0; i<4; i++)
		{
			if((n.n[i].a > 0.5)) //neighbour active
			{
				//set fill color if own color matches fill color
				if(Equals(col, FillCol)) 
				{
					col = FillCol;
				}
			}
		}
	}
	
    return col;
}

// --------------------------------------------------------------------------------------------------
// TECHNIQUES:
// --------------------------------------------------------------------------------------------------

technique TFill
{
    pass P0
    {
        //Wrap0 = U;  // useful when mesh is round like a sphere
        VertexShader = compile vs_1_1 VS();
        PixelShader  = compile ps_2_0 PS();
    }
}

technique TFixedFunction
{
    pass P0
    {
        //transforms
        WorldTransform[0]   = (tW);
        ViewTransform       = (tV);
        ProjectionTransform = (tP);

        //texturing
        Sampler[0] = (Samp);
        TextureTransform[0] = (tTex);
        TexCoordIndex[0] = 0;
        TextureTransformFlags[0] = COUNT2;
        //Wrap0 = U;  // useful when mesh is round like a sphere
        
        Lighting       = FALSE;

        //shaders
        VertexShader = NULL;
        PixelShader  = NULL;
    }
}
